use std::{env, fs};

use camino::Utf8PathBuf;
use chrono::{Datelike, Utc};
use clap::Parser;
use color_eyre::eyre::{bail, eyre, OptionExt, Result};
use log::{debug, error, info};
use regex::Regex;
use reqwest::StatusCode;
use serde::{Deserialize, Serialize};

#[derive(Debug, Clone, Serialize, Deserialize)]
struct Task {
    title: String,
    created: String,
    description: String,
    done: bool,
}

fn default_title() -> String {
    let now = Utc::now();
    format!(
        "Teamtreffen {:02}.{:02}.{}",
        now.day(),
        now.month(),
        now.year()
    )
}

fn default_output() -> Utf8PathBuf {
    let now = Utc::now();
    format!(
        "{:02}.{:02}.{:02}.md",
        now.year() % 100,
        now.month(),
        now.day()
    )
    .into()
}

/// Simple program to fill a markdown template with today's date and bulletpoints from the
/// corresponding vikunja board.
#[derive(Debug, Parser)]
#[command(version, about)]
struct Args {
    /// Markdown template to fill with today's points
    #[arg(short = 'i', long, default_value = "template.md")]
    template: Utf8PathBuf,
    /// The url to make requests to
    #[arg(short, long, default_value = "https://vikunja.luhbots-hannover.de")]
    url: String,
    /// Title of the task on vikunja
    #[arg(short, long, default_value_t = default_title())]
    title: String,
    /// File to write the filled template to
    #[arg(short, long, default_value_t = default_output())]
    output: Utf8PathBuf,
}

fn main() -> Result<()> {
    color_eyre::install()?;
    env_logger::init();

    debug!("parsing arguments");
    let args = Args::parse();

    if !args.template.exists() {
        error!("The template {} does not exist", args.template);
        return Err(eyre!("No such file or directory: {}", args.template));
    }

    debug!("creating client");
    let client = reqwest::blocking::Client::new();
    info!("requesting tasks from vikunja");
    let response = client
        .get(format!("{}/api/v1/tasks/all", args.url))
        .header(
            reqwest::header::AUTHORIZATION,
            format!("Bearer {}", env::var("VIKUNJA_TOKEN")?),
        )
        .query(&[
            // search for tasks containing "Teamtreffen"
            ("s", args.title.as_str()),
            // sort by creation data
            ("sort_by", "created"),
            // sort in descending order (newest first)
            ("order_by", "desc"),
        ])
        .send()?;
    debug!("got response: {response:?}");

    if response.status() != StatusCode::OK {
        error!("Request to Vikunja failed!");
        let status = response.status();
        let body_text = response.text()?;
        info!("{}", body_text);
        bail!("Vikunja request failed {}: {}", status, body_text)
    }

    info!("searching for the correct task");
    let tasks = response.json::<Vec<Task>>()?;
    debug!("got tasks: {tasks:#?}");
    let task = tasks
        .into_iter()
        .find(|task| task.title == args.title)
        .ok_or_eyre("Unable to find Task for today's Team meeting")?;

    info!("extracting bullet points");
    let re = Regex::new("<p>([^<]+)</p>")?;
    let points = re
        .captures_iter(&task.description)
        .map(|c| c.extract::<1>().1[0])
        .collect::<Vec<_>>();

    info!("reading template from {}", args.template);
    let template = fs::read_to_string(args.template)?;

    info!("replacing dates in template");
    let now = Utc::now();
    let mut template = template.replace(
        "dd.mm.yyyy",
        format!("{:02}.{:02}.{}", now.day(), now.month(), now.year()).as_str(),
    );

    info!("adding bullet points to the template");
    template.extend(
        points
            .iter()
            .filter(|point| **point != "Rasender Report" && **point != "Was ging die Woche?")
            .map(|point| format!("\n# {point}\n")),
    );

    info!("saving to {}", args.output);
    fs::write(args.output, template)?;

    Ok(())
}
